#!/usr/bin/env python

import pytest
import numpy as np
import pandas as pd
from gbstats.gbstats import (
    check_srm,
    get_adjusted_stats,
    process_user_rows,
    process_metric_rows,
    run_analysis,
)
from functools import partial

DECIMALS = 9
round_ = partial(np.round, decimals=DECIMALS)


def test_srm():
    p = check_srm([1000, 1200], [0.5, 0.5])
    assert round_(p) == 0.000020079


def test_adjusted_stats():
    adjusted = get_adjusted_stats(5, 3, 1000, 2000, False, "revenue")
    print(adjusted)
    assert adjusted["users"] == 2000
    assert adjusted["mean"] == 2.5
    assert round_(adjusted["stddev"]) == 3.278852762
    assert adjusted["total"] == 5000


def test_adjusted_stats_binomial():
    adjusted = get_adjusted_stats(1, 0, 1000, 2000, False, "binomial")
    print(adjusted)
    assert adjusted["users"] == 2000
    assert adjusted["mean"] == 1
    assert round_(adjusted["stddev"]) == 0
    assert adjusted["total"] == 1000


def test_adjusted_stats_ignore_nulls():
    adjusted = get_adjusted_stats(5, 3, 1000, 2000, True, "revenue")
    assert adjusted["users"] == 1000
    assert adjusted["mean"] == 5
    assert adjusted["stddev"] == 3
    assert adjusted["total"] == 5000


def test_process_users():
    vars = {"zero": 0, "one": 1}
    rows = pd.DataFrame(
        [{"variation": "one", "users": 120}, {"variation": "zero", "users": 100}]
    )
    users, unknown_variations = process_user_rows(rows, vars)

    assert users == [100, 120]
    assert unknown_variations == []


def test_process_users_unknown_vars():
    var_id_map = {"zero": 0, "one": 1}
    rows = pd.DataFrame(
        [{"variation": "one", "users": 120}, {"variation": "zeros", "users": 100}]
    )
    users, unknown_variations = process_user_rows(rows, var_id_map)

    assert users == [0, 120]
    assert unknown_variations == ["zeros"]


def test_process_metrics():
    rows = pd.DataFrame(
        [
            {"variation": "one", "count": 120, "mean": 2.5, "stddev": 1},
            {"variation": "zero", "count": 100, "mean": 2.7, "stddev": 1.1},
        ]
    )
    var_id_map = {"zero": 0, "one": 1}
    users = [1000, 1010]

    res = process_metric_rows(rows, var_id_map, users, False, "revenue")
    assert res.loc[0].at["users"] == 1000
    assert res.loc[0].at["count"] == 100
    assert res.loc[0].at["mean"] == 0.27
    assert round_(res.loc[0].at["stddev"]) == 0.881286938


def test_process_metrics_ignore_nulls():
    rows = pd.DataFrame(
        [
            {"variation": "one", "count": 120, "mean": 2.5, "stddev": 1},
            {"variation": "zero", "count": 100, "mean": 2.7, "stddev": 1.1},
        ]
    )
    var_id_map = {"zero": 0, "one": 1}
    users = [1000, 1010]

    res = process_metric_rows(rows, var_id_map, users, True, "revenue")
    assert res.loc[0].at["users"] == 100
    assert res.loc[0].at["count"] == 100
    assert res.loc[0].at["mean"] == 2.7
    assert round_(res.loc[0].at["stddev"]) == 1.1


def test_binomial_analysis():
    metric = pd.DataFrame(
        [
            {"users": 1000, "count": 120, "mean": 1, "stddev": 0, "total": 120},
            {"users": 1024, "count": 128, "mean": 1, "stddev": 0, "total": 128},
            {"users": 1000, "count": 102, "mean": 1, "stddev": 0, "total": 102},
        ]
    )
    var_names = ["Control", "Variation 1", "Variation 2"]
    res = run_analysis(metric, var_names, "binomial", False)

    baseline = res.loc[0]
    var1 = res.loc[1]
    var2 = res.loc[2]

    assert baseline.at["variation"] == "Control"
    assert baseline.at["conversion_rate"] == 0.12
    assert baseline.at["chance_to_beat_control"] == None
    assert round_(baseline.at["risk_of_choosing"]) == 0.069118343
    assert baseline.at["percent_change"] == None

    assert var1.at["variation"] == "Variation 1"
    assert var1.at["conversion_rate"] == 0.125
    assert round_(var1.at["chance_to_beat_control"]) == 0.633751254
    assert round_(var1.at["risk_of_choosing"]) == 0.029338254
    assert round_(var1.at["percent_change"]) == 0.041432724

    assert var2.at["variation"] == "Variation 2"
    assert var2.at["conversion_rate"] == 0.102
    assert round_(var2.at["chance_to_beat_control"]) == 0.100849049
    assert round_(var2.at["risk_of_choosing"]) == 0.182688464
    assert round_(var2.at["percent_change"]) == -0.149376661


def test_gaussian_analysis():
    metric = pd.DataFrame(
        [
            {"users": 1000, "count": 120, "mean": 1.3, "stddev": 1, "total": 156},
            {"users": 1024, "count": 128, "mean": 1.29, "stddev": 0.9, "total": 165.12},
            {"users": 1000, "count": 102, "mean": 1.4, "stddev": 1.1, "total": 142.8},
        ]
    )
    var_names = ["Control", "Variation 1", "Variation 2"]
    res = run_analysis(metric, var_names, "duration", True)

    baseline = res.loc[0]
    var1 = res.loc[1]
    var2 = res.loc[2]

    assert baseline.at["variation"] == "Control"
    assert baseline.at["per_user"] == 0.156
    assert baseline.at["chance_to_beat_control"] == None
    assert round_(baseline.at["risk_of_choosing"]) == 0.138620458
    assert baseline.at["percent_change"] == None

    assert var1.at["variation"] == "Variation 1"
    assert var1.at["per_user"] == 0.16125
    assert round_(var1.at["chance_to_beat_control"]) == 0.593436958
    assert round_(var1.at["risk_of_choosing"]) == 0.076604954
    assert round_(var1.at["percent_change"]) == -0.007692308

    assert var2.at["variation"] == "Variation 2"
    assert round_(var2.at["per_user"]) == 0.1428
    assert round_(var2.at["chance_to_beat_control"]) == 0.016533047
    assert round_(var2.at["risk_of_choosing"]) == 0.702254931
    assert round_(var2.at["percent_change"]) == 0.076923077



from abc import ABC, abstractmethod
from warnings import warn
import numpy as np
from scipy.stats import beta, norm, rv_continuous
from scipy.special import digamma, polygamma, roots_hermitenorm
from .orthogonal import roots_sh_jacobi

EPSILON = 1e-04


class BayesABDist(ABC):
    dist: rv_continuous

    @staticmethod
    @abstractmethod
    def posterior(prior, data):
        """
        :type prior: Iterable
        :type data: Iterable
        :rtype: Tuple[ndarray, ndarray]
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def moments(par1, par2, log=False):
        """
        :type par1: float or ndarray
        :type par2: float or ndarray
        :type log: bool
        :rtype: Tuple[float or ndarray, float or ndarray]
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def gq(n, par1, par2):
        """
        :type n: int
        :type par1: float
        :type par2: float
        :rtype: Tuple[ndarray, ndarray]
        """
        raise NotImplementedError

    # todo: @vectorize
    @classmethod
    def risk(cls, a_par1, a_par2, b_par1, b_par2, n=24):
        """
        :type a_par1: float
        :type a_par2: float
        :type b_par1: float
        :type b_par2: float
        :type n: int
        :rtype: ndarray
        """
        a_nodes, a_weights = cls.gq(n, a_par1, a_par2)
        b_nodes, b_weights = cls.gq(n, b_par1, b_par2)

        gq = sum(a_nodes * cls.dist.cdf(a_nodes, b_par1, b_par2) * a_weights) + sum(
            b_nodes * cls.dist.cdf(b_nodes, a_par1, a_par2) * b_weights
        )
        out = gq - cls.dist.mean((a_par1, b_par1), (a_par2, b_par2))

        return out


class Beta(BayesABDist):
    dist = beta

    @staticmethod
    def posterior(prior, data):
        a = prior[0] + data[0]
        b = prior[1] + data[1] - data[0]
        return a, b

    @staticmethod
    def moments(par1, par2, log=False):
        if np.sum(par2 < 0) + np.sum(par1 < 0):
            raise RuntimeError("params of beta distribution cannot be negative")

        if log:
            mean = digamma(par1) - digamma(par1 + par2)
            var = polygamma(1, par1) - polygamma(1, par1 + par2)
        else:
            mean = par1 / (par1 + par2)
            var = par1 * par2 / (np.power(par1 + par2, 2) * (par1 + par2 + 1))
        return mean, var

    @staticmethod
    def gq(n, par1, par2):
        x, w = roots_sh_jacobi(int(n), par1 + par2 - 1, par1, False)
        return x, w


class Norm(BayesABDist):
    dist = norm

    @staticmethod
    def posterior(prior, data):
        inv_var_0 = prior[2] / np.power(prior[1], 2)
        inv_var_d = data[2] / np.power(data[1], 2)
        var = 1 / (inv_var_0 + inv_var_d)

        loc = var * (inv_var_0 * prior[0] + inv_var_d * data[0])
        scale = np.sqrt(var)
        return loc, scale

    @staticmethod
    def moments(par1, par2, log=False):
        if np.sum(par2 < 0):
            raise RuntimeError("got negative standard deviation.")

        if log:
            if np.sum(par1 <= 0):
                raise RuntimeError("got mu <= 0. cannot use log approximation.")

            max_prob = np.max(norm.cdf(0, par1, par2))
            if max_prob > EPSILON:
                warn(
                    f"probability of being negative is higher than {EPSILON} (={max_prob}). "
                    f"log approximation is in-exact",
                    RuntimeWarning,
                )

            mean = np.log(par1)
            var = np.power(par2 / par1, 2)
        else:
            mean = par1
            var = np.power(par2, 2)
        return mean, var

    @staticmethod
    def gq(n, par1, par2):
        if par2 <= 0:
            raise RuntimeError("got negative standard deviation.")

        x, w, m = roots_hermitenorm(int(n), True)
        x = par2 * x + par1
        w /= m
        return x, w



import numpy as np
from scipy.stats import norm
from .dists import Beta, Norm


"""
Medium article inspiration:
    https://towardsdatascience.com/how-to-do-bayesian-a-b-testing-fast-41ee00d55be8

Original code:
    https://github.com/itamarfaran/public-sandbox/tree/master/bayesian_blog
"""


BETA_PRIOR = 1, 1
NORM_PRIOR = 0, 1, 0


def binomial_ab_test(x_a, n_a, x_b, n_b, ccr=0.05):
    alpha_a, beta_a = Beta.posterior(BETA_PRIOR, [x_a, n_a])
    alpha_b, beta_b = Beta.posterior(BETA_PRIOR, [x_b, n_b])

    mean_a, var_a = Beta.moments(alpha_a, beta_a, log=True)
    mean_b, var_b = Beta.moments(alpha_b, beta_b, log=True)

    mean_diff = mean_b - mean_a
    std_diff = np.sqrt(var_a + var_b)

    chance_to_win = norm.sf(0, mean_diff, std_diff)
    expected = np.exp(mean_diff) - 1
    ci = np.exp(norm.ppf([ccr / 2, 1 - ccr / 2], mean_diff, std_diff)) - 1
    risk_beta = Beta.risk(alpha_a, beta_a, alpha_b, beta_b)

    output = {
        "chance_to_win": chance_to_win,
        "expected": expected,
        "ci": ci.tolist(),
        "uplift": {"dist": "lognormal", "mean": mean_diff, "stddev": std_diff},
        "risk": risk_beta.tolist(),
    }

    return output


def gaussian_ab_test(m_a, s_a, n_a, m_b, s_b, n_b, ccr=0.05):
    mu_a, sd_a = Norm.posterior(NORM_PRIOR, [m_a, s_a, n_a])
    mu_b, sd_b = Norm.posterior(NORM_PRIOR, [m_b, s_b, n_b])

    mean_a, var_a = Norm.moments(mu_a, sd_a, log=True)
    mean_b, var_b = Norm.moments(mu_b, sd_b, log=True)

    mean_diff = mean_b - mean_a
    std_diff = np.sqrt(var_a + var_b)

    chance_to_win = norm.sf(0, mean_diff, std_diff)
    expected = np.exp(mean_diff) - 1
    ci = np.exp(norm.ppf([ccr / 2, 1 - ccr / 2], mean_diff, std_diff)) - 1
    risk_norm = Norm.risk(mu_a, sd_a, mu_b, sd_b)

    output = {
        "chance_to_win": chance_to_win,
        "expected": expected,
        "ci": ci.tolist(),
        "uplift": {"dist": "lognormal", "mean": mean_diff, "stddev": std_diff},
        "risk": risk_norm.tolist(),
    }

    return output



import pandas as pd
import math
from .bayesian.main import binomial_ab_test, gaussian_ab_test
from scipy.stats.distributions import chi2


# Adjust metric stats to account for unconverted users
def get_adjusted_stats(x, sx, c, n, ignore_nulls=False, type="binomial"):
    if type == "binomial":
        return {"users": n, "count": c, "mean": x, "stddev": sx, "total": c}
    # Ignore unconverted users
    elif ignore_nulls:
        return {"users": c, "count": c, "mean": x, "stddev": sx, "total": c * x}
    # Add in unconverted users and correct the mean/stddev
    else:
        mean = (x * c) / n
        varx = sx ** 2
        # From https://math.stackexchange.com/questions/2971315/how-do-i-combine-standard-deviations-of-two-groups
        stddev = math.sqrt(
            ((c - 1) * varx) / (n - 1) + (c * (n - c) * (x ** 2)) / (n * (n - 1))
        )

        return {"users": n, "count": c, "mean": mean, "stddev": stddev, "total": x * c}


# Transform raw SQL result for metrics into a list of stats per variation
def process_metric_rows(rows, var_id_map, users, ignore_nulls=False, type="binomial"):
    stats = [{"users": 0, "count": 0, "mean": 0, "stddev": 0, "total": 0}] * len(
        var_id_map.keys()
    )
    for row in rows.itertuples(index=False):
        id = str(row.variation)
        if id in var_id_map:
            variation = var_id_map[id]
            stats[variation] = get_adjusted_stats(
                x=row.mean,
                sx=row.stddev,
                c=row.count,
                n=users[variation],
                ignore_nulls=ignore_nulls,
                type=type,
            )
    return pd.DataFrame(stats)


# Transform raw SQL result for users into a list of num_users per variation
def process_user_rows(rows, var_id_map):
    users = [0] * len(var_id_map.keys())
    unknown_var_ids = []
    for row in rows.itertuples(index=False):
        id = str(row.variation)
        if id in var_id_map:
            variation = var_id_map[id]
            users[variation] = row.users
        else:
            unknown_var_ids.append(id)
    return users, unknown_var_ids


# Run A/B test analysis for a metric
def run_analysis(metric, var_names, type="binomial", inverse=False):
    vars = iter(metric.itertuples(index=False))
    baseline = next(vars)

    # baseline users, mean, count, stddev, and value
    n_a = baseline.users
    m_a = baseline.mean
    x_a = baseline.count
    s_a = baseline.stddev
    v_a = baseline.total

    cr_a = v_a / n_a

    ret = pd.DataFrame(
        [
            {
                "variation": var_names[0],
                "users": n_a,
                "total": v_a,
                "per_user": cr_a,
                "chance_to_beat_control": None,
                "risk_of_choosing": None,
                "percent_change": None,
                "uplift_dist": None,
                "uplift_mean": None,
                "uplift_stddev": None,
            }
        ]
    )

    baseline_risk = 0
    for i, row in enumerate(vars):
        # variation users, mean, count, stddev, and value
        n_b = row.users
        m_b = row.mean
        x_b = row.count
        s_b = row.stddev
        v_b = row.total
        cr_b = v_b / n_b

        if type == "binomial":
            res = binomial_ab_test(x_a, n_a, x_b, n_b)
        else:
            res = gaussian_ab_test(m_a, s_a, n_a, m_b, s_b, n_b)

        # Flip risk and chance to win for inverse metrics
        risk0 = res["risk"][0] if not inverse else res["risk"][1]
        risk1 = res["risk"][1] if not inverse else res["risk"][0]
        ctw = res["chance_to_win"] if not inverse else 1 - res["chance_to_win"]

        # Turn risk into relative risk
        risk0 = risk0 / cr_b
        risk1 = risk1 / cr_b

        if risk0 > baseline_risk:
            baseline_risk = risk0

        s = pd.Series(
            {
                "variation": var_names[i + 1],
                "users": n_b,
                "total": v_b,
                "per_user": cr_b,
                "chance_to_beat_control": ctw,
                "risk_of_choosing": risk1,
                "percent_change": res["expected"],
                "uplift_dist": res["uplift"]["dist"],
                "uplift_mean": res["uplift"]["mean"],
                "uplift_stddev": res["uplift"]["stddev"],
            }
        )

        ret = ret.append(s, ignore_index=True)

    ret.at[0, "risk_of_choosing"] = baseline_risk

    # Rename columns for binomial metrics
    if type == "binomial":
        ret.rename(
            columns={"total": "conversions", "per_user": "conversion_rate"},
            inplace=True,
        )

    return ret


def check_srm(users, weights):
    # Convert count of users into ratios
    total_observed = sum(users)
    if not total_observed:
        return 1

    x = 0
    for i, o in enumerate(users):
        e = weights[i] * total_observed
        x = x + ((o - e) ** 2) / e

    return chi2.sf(x, len(users) - 1)



import numpy as np
from scipy import linalg
from scipy.special import betaln, eval_jacobi


"""
see scipy.special.orthogonal.
copied source code and implemented log trick to avoid OverflowError
"""


def _gen_roots_and_weights(n, log_mu0, an_func, bn_func, f, df, mu):
    """
    see _gen_roots_and_weights in scipy.special.orthogonal
    """
    k = np.arange(n, dtype="d")
    c = np.zeros((2, n))
    c[0, 1:] = bn_func(k[1:])
    c[1, :] = an_func(k)
    x = linalg.eigvals_banded(c, overwrite_a_band=True)

    # improve roots by one application of Newton's method
    y = f(n, x)
    dy = df(n, x)
    x -= y / dy

    fm = f(n - 1, x)
    fm /= np.abs(fm).max()
    dy /= np.abs(dy).max()
    w = 1.0 / (fm * dy)

    log_w = np.log(w) + log_mu0 - np.log(w.sum())

    if mu:
        return x, log_w, log_mu0
    else:
        return x, log_w


def roots_jacobi(n, alpha, beta):
    """
    Gauss-Jacobi quadrature.
    see scipy.special.root_jacobi
    """

    def an_func(k):
        if a + b == 0.0:
            return np.where(k == 0, (b - a) / (2 + a + b), 0.0)
        return np.where(
            k == 0,
            (b - a) / (2 + a + b),
            (b * b - a * a) / ((2.0 * k + a + b) * (2.0 * k + a + b + 2)),
        )

    def bn_func(k):
        return (
            2.0
            / (2.0 * k + a + b)
            * np.sqrt((k + a) * (k + b) / (2 * k + a + b + 1))
            * np.where(k == 1, 1.0, np.sqrt(k * (k + a + b) / (2.0 * k + a + b - 1)))
        )

    def f(n, x):
        return eval_jacobi(n, a, b, x)

    def df(n, x):
        return 0.5 * (n + a + b + 1) * eval_jacobi(n - 1, a + 1, b + 1, x)

    m = int(n)
    if n < 1 or n != m:
        raise ValueError("n must be a positive integer.")
    if alpha <= -1 or beta <= -1:
        raise ValueError("alpha and beta must be greater than -1.")

    log_mu0 = (alpha + beta + 1) * np.log(2.0) + betaln(alpha + 1, beta + 1)
    a = alpha
    b = beta

    return _gen_roots_and_weights(m, log_mu0, an_func, bn_func, f, df, True)


def roots_sh_jacobi(n, p1, q1, mu=False):
    """
    Gauss-Jacobi (shifted) quadrature.
    see scipy.special.roots_sh_jacobi
    used the log trick to integrate over large values of a,b
    """
    if (p1 - q1) <= -1 or q1 <= 0:
        raise ValueError(
            "(p - q) must be greater than -1, and q must be greater than 0."
        )
    x, log_w, log_m = roots_jacobi(n, p1 - q1, q1 - 1)
    x = (x + 1) / 2
    w = np.exp(log_w - log_m)
    if mu:
        return x, w, 1
    else:
        return x, w



from functools import partial
from unittest import TestCase, main as unittest_main
import numpy as np
from gbstats.bayesian.main import binomial_ab_test, gaussian_ab_test

DECIMALS = 5
round_ = partial(np.round, decimals=DECIMALS)


class TestBinom(TestCase):
    def test_binomial_ab_test(self):
        result = binomial_ab_test(49, 100, 51, 100)
        expected = {
            "chance_to_win": 0.61052,
            "expected": 0.0404,
            "ci": None,
            "uplift": None,
            "risk": [0.03872, 0.01912],
        }

        for key in expected.keys():
            ex = expected[key]
            res = result[key]

            if ex is None:
                continue
            res = [round_(x) for x in res] if isinstance(res, list) else round_(res)
            ex = [round_(x) for x in ex] if isinstance(ex, list) else round_(ex)
            self.assertEqual(res, ex)


class TestNorm(TestCase):
    def test_gaussian_ab_test(self):
        result = gaussian_ab_test(10, 0.5, 10, 10.5, 1, 10)
        expected = {
            "chance_to_win": 0.92427,
            "expected": 0.05,
            "ci": None,
            "uplift": None,
            "risk": [0.51256, 0.01256],
        }

        for key in expected.keys():
            ex = expected[key]
            res = result[key]

            if ex is None:
                continue
            res = [round_(x) for x in res] if isinstance(res, list) else round_(res)
            ex = [round_(x) for x in ex] if isinstance(ex, list) else round_(ex)
            self.assertEqual(res, ex)


if __name__ == "__main__":
    unittest_main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gbstats', 'gbstats.bayesian', 'tests']

package_data = \
{'': ['*']}

extras_require = \
{'dev': ['tox>=3.20.1,<4.0.0',
         'virtualenv>=20.2.2,<21.0.0',
         'pip>=20.3.1,<21.0.0',
         'twine>=3.3.0,<4.0.0',
         'toml>=0.10.2,<0.11.0'],
 'test': ['black==20.8b1',
          'isort==5.6.4',
          'flake8==3.8.4',
          'flake8-docstrings>=1.6.0,<2.0.0',
          'pytest==6.1.2',
          'pytest-cov==2.10.1']}

setup_kwargs = {
    'name': 'gbstats',
    'version': '0.1.1',
    'description': 'Stats engine for GrowthBook, the open source A/B testing platform.',
    'long_description': '# GrowthBook Stats\n\nThe stats engine for GrowthBook, the open source A/B testing platform.\n\n## Installation\n\n```\npip install gbstats\n```\n\n## Usage\n\n```python\nimport gbstats\n```\n',
    'author': 'Jeremy Dorn',
    'author_email': 'jeremy@growthbook.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/growthbook/growthbook/packages/stats',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0',
}


setup(**setup_kwargs)



from functools import partial
from unittest import TestCase, main as unittest_main

import numpy as np
import pandas as pd
from scipy.special import digamma
from scipy.stats import beta, norm
from gbstats.bayesian.dists import Beta, Norm


DECIMALS = 5
round_ = partial(np.round, decimals=DECIMALS)


def roundsum(x, decimals=DECIMALS):
    return np.round(np.sum(x), decimals=decimals)


class TestBeta(TestCase):
    def test_posterior(self):
        prior = 1, 1
        data = 1, 2
        result = Beta.posterior(prior, data)
        outcome = (2, 2)

        for res, out in zip(result, outcome):
            self.assertEqual(res, out)

        prior = 1, 1
        data = pd.Series([1, 10]), pd.Series([2, 20])
        result = Beta.posterior(prior, data)
        outcome = pd.Series([2, 11]), pd.Series([2, 11])

        for res, out in zip(result, outcome):
            pd.testing.assert_series_equal(res, out)

        prior = pd.Series([1, 2]), pd.Series([1, 3])
        data = pd.Series([1, 10]), pd.Series([2, 20])
        result = Beta.posterior(prior, data)
        outcome = pd.Series([2, 12]), pd.Series([2, 13])

        for res, out in zip(result, outcome):
            pd.testing.assert_series_equal(res, out)

    def test_moments(self):
        pars = 12, 745
        result = Beta.moments(*pars)
        expected = beta.mean(*pars), beta.var(*pars)
        for res, out in zip(result, expected):
            self.assertEqual(round_(res), round_(out))

        pars = 12, 745
        result = Beta.moments(*pars, log=True)
        mean = beta.expect(np.log, pars)
        var = beta.expect(lambda x: np.log(x) ** 2, pars) - mean ** 2
        expected = mean, var
        for res, out in zip(result, expected):
            self.assertEqual(round_(res), round_(out))

        pars = np.array([12, 745]), np.array([745, 12])
        result = Beta.moments(*pars)
        expected = beta.mean(*pars), beta.var(*pars)
        for res, out in zip(result, expected):
            np.testing.assert_array_almost_equal(res, out)

        self.assertRaises(RuntimeError, Beta.moments, 1, -1)
        self.assertRaises(RuntimeError, Beta.moments, -1, 1)
        self.assertRaises(RuntimeError, Beta.moments, -1, -1)

    def test_gq(self):
        test_cases = zip([10, 100, 500, 1000, 10000], [10000, 1000, 500, 100, 10])
        for a, b in test_cases:
            x, w = Beta.gq(24, a, b)
            for p in range(8):
                self.assertEqual(roundsum(x ** p * w), roundsum(beta.moment(p, a, b)))
            self.assertEqual(
                roundsum(np.log(x) * w), roundsum(digamma(a) - digamma(a + b))
            )


class TestNorm(TestCase):
    def test_posterior(self):
        prior = 0, 1, 10
        data = 12, 1, 10
        result = Norm.posterior(prior, data)
        outcome = (6, np.sqrt(1 / 20))

        for res, out in zip(result, outcome):
            self.assertEqual(res, out)

        prior = 0, 1, 10
        data = pd.Series([12, 100]), pd.Series([1, np.sqrt(2)]), pd.Series([10, 20])
        result = Norm.posterior(prior, data)
        outcome = pd.Series([6.0, 50.0]), pd.Series([np.sqrt(1 / 20), np.sqrt(1 / 20)])

        for res, out in zip(result, outcome):
            pd.testing.assert_series_equal(res, out)

        prior = pd.Series([0, 100]), pd.Series([1, np.sqrt(2)]), pd.Series([10, 20])
        data = pd.Series([12, 100]), pd.Series([1, np.sqrt(2)]), pd.Series([10, 20])
        result = Norm.posterior(prior, data)
        outcome = pd.Series([6.0, 100.0]), pd.Series([np.sqrt(1 / 20), np.sqrt(1 / 20)])

        for res, out in zip(result, outcome):
            pd.testing.assert_series_equal(res, out)

    def test_moments(self):
        pars = 10, 100
        result = Norm.moments(*pars)
        expected = norm.mean(*pars), norm.var(*pars)
        for res, out in zip(result, expected):
            self.assertEqual(round_(res), round_(out))

        pars = 100, 10
        result = Norm.moments(*pars, log=True)
        expected = np.log(100), (10 / 100) ** 2
        for res, out in zip(result, expected):
            self.assertEqual(round_(res), round_(out))

        pars = np.array([10, 100]), np.array([100, 10])
        result = Norm.moments(*pars)
        expected = norm.mean(*pars), norm.var(*pars)
        for res, out in zip(result, expected):
            np.testing.assert_array_almost_equal(res, out)

        self.assertWarns(RuntimeWarning, Norm.moments, 0.1, 1, log=True)
        self.assertRaises(RuntimeError, Norm.moments, 0, 1, log=True)
        self.assertRaises(RuntimeError, Norm.moments, -1, 1, log=True)
        self.assertRaises(RuntimeError, Norm.moments, 1, -1)

    def test_gq(self):
        test_cases = zip([0, -2, 2, 10], [0.01, 1, 4, 0.0001])
        for loc, scale in test_cases:
            x, w = Norm.gq(24, loc, scale)
            for p in range(8):
                self.assertEqual(
                    roundsum(x ** p * w), roundsum(norm.moment(p, loc, scale))
                )

        self.assertRaises(RuntimeError, Norm.gq, 24, 0, 0)
        self.assertRaises(RuntimeError, Norm.gq, 24, 0, -1)


if __name__ == "__main__":
    unittest_main()



from .gbstats import process_metric_rows, process_user_rows, run_analysis, check_srm
import pandas as pd
import nbformat
from nbformat import v4 as nbf
from nbformat.v4.nbjson import from_dict


def code_cell_df(df, source=""):
    return nbf.new_code_cell(
        source=source,
        outputs=[
            nbf.new_output(
                output_type="execute_result",
                data={"text/plain": df.to_string(), "text/html": df.to_html()},
            )
        ],
    )


def code_cell_plain(source="", text=""):
    return nbf.new_code_cell(
        source=source,
        outputs=[
            nbf.new_output(output_type="execute_result", data={"text/plain": text})
        ],
    )


def create_notebook(
    url="",
    hypothesis="",
    name="",
    var_id_map={},
    var_names=[],
    weights=[],
    run_query="",
    users_sql="",
    user_rows=None,
    metrics=[],
):
    users, unknown_var_ids = process_user_rows(user_rows, var_id_map)
    srm_p = check_srm(users, weights)

    cells = [
        # Intro
        nbf.new_markdown_cell(
            f"# {name}\n"
            f"[View on GrowthBook]({url})\n\n"
            f"**Hypothesis:** {hypothesis}\n\n"
            "## Notebook Setup"
        ),
        # Notebook Setup
        nbf.new_code_cell(
            "from gbstats.gbstats import process_user_rows, "
            "check_srm, process_metric_rows, run_analysis\n\n"
            "# Mapping of variation id to index\n"
            f"var_id_map = {str(var_id_map)}\n\n"
            "# Display names of variations\n"
            f"var_names = {str(var_names)}\n\n"
            "# Expected traffic split between variations\n"
            f"weights = {str(weights)}"
        ),
        # runQuery definition
        nbf.new_code_cell("# User defined runQuery function\n" f"{run_query}"),
        # Users in Experiment
        nbf.new_markdown_cell("## Users in Experiment\n" "### Query"),
        # Users Query
        code_cell_df(
            user_rows,
            source=(
                f'users_sql = """{users_sql}"""\n'
                "user_rows = runQuery(users_sql)\n"
                "display(user_rows)"
            ),
        ),
        nbf.new_markdown_cell("### Data Quality Checks"),
        code_cell_plain(
            source=(
                "# Process raw user rows\n"
                "users, unknown_var_ids = process_user_rows(user_rows, var_id_map)\n\n"
                "# Users in each variation\n"
                'print("Users in each variation:", users)\n\n'
                "# Any variation ids returned from the query that we were not expecting\n"
                'print("Unknown variation ids:", unknown_var_ids)'
            ),
            text=(
                f"Users in each variation:"
                f"{str(users)}\n"
                "Unknown variation ids:"
                f"{str(unknown_var_ids)}"
            ),
        ),
        code_cell_plain(
            source=(
                "# Sample Ratio Mismatch (SRM) Check\n"
                "srm_p = check_srm(users, weights)\n\n"
                'print("SRM P-value:", srm_p)\n\n'
                "if srm_p < 0.001:\n"
                '    print("***WARNING: Sample Ratio Mismatch Detected***")\n'
                "else:\n"
                '    print("Ok, no SRM detected")'
            ),
            text=(
                f"SRM P-value: {srm_p}\n"
                f'{"***WARNING: Sample Ratio Mismatch Detected***" if srm_p < 0.001 else "Ok, no SRM detected"}'
            ),
        ),
    ]

    for i, metric in enumerate(metrics):
        name = metric["name"]
        cells.append(nbf.new_markdown_cell(f"## Metric - {name}\n" "### Query"))
        sql = metric["sql"]
        cells.append(
            code_cell_df(
                df=metric["rows"],
                source=(
                    "# Get aggregate metric values per variation\n"
                    f'm{i}_sql = """{sql}"""\n\n'
                    f"m{i}_rows = runQuery(m{i}_sql)\n"
                    f"display(m{i}_rows)"
                ),
            )
        )
        cells.append(nbf.new_markdown_cell("### Preparation"))

        type = metric["type"]
        ignore_nulls = metric["ignore_nulls"]
        inverse = metric["inverse"]
        processed = process_metric_rows(
            rows=metric["rows"],
            var_id_map=var_id_map,
            users=users,
            ignore_nulls=ignore_nulls,
            type=type,
        )

        cells.append(
            code_cell_df(
                df=processed,
                source=(
                    f"m{i} = process_metric_rows(\n"
                    f"    rows=m{i}_rows,\n"
                    f"    var_id_map=var_id_map,\n"
                    f"    users=users,\n"
                    f"    ignore_nulls={ignore_nulls},\n"
                    f'    type="{type}"\n'
                    f")\n"
                    f"display(m{i})"
                ),
            )
        )

        cells.append(nbf.new_markdown_cell("### Result"))

        result = run_analysis(
            metric=processed, var_names=var_names, type=type, inverse=inverse
        )

        cells.append(
            code_cell_df(
                df=result,
                source=(
                    f"m{i}_result = run_analysis(\n"
                    f"    metric=m{i},\n"
                    f"    var_names=var_names,\n"
                    f'    type="{type}",\n'
                    f"    inverse={inverse}\n"
                    f")\n"
                    f"display(m{i}_result)"
                ),
            )
        )

    nb = nbf.new_notebook(
        metadata=from_dict(
            {
                "kernelspec": {
                    "display_name": "Python 3",
                    "language": "python",
                    "name": "python3",
                },
                "language_info": {
                    "name": "python",
                    "codemirror_mode": {"name": "ipython", "version": 3},
                    "version": "3.7.4",
                    "file_extension": ".py",
                    "mimetype": "text/x-python",
                    "nbconvert_exporter": "python",
                    "pygments_lexer": "ipython3",
                },
            }
        ),
        cells=cells,
    )

    return nbformat.writes(
        nb=nb,
    )
